import poser

#This EZskin3 plugin was made to use the new capabilities of the Cyclenodes used with the new superfly render of poser11 and poser11pro.
#the intention was to make an Superfly overlay shader on the existing Firefly shaders so it still could be used with firefly.
#In other words, it will keep your Firefly shader intact.
#Made by Thijs Broekhuijsen AKA Bopperthijs, Bopperman.
#Updated 23-11-2016

# Global variables
EZBOOL, EZBUTTON, EZINT, EZCOLOUR, EZCOLOURP, EZFLOAT, EZCHOICE, EZSELECT, EZISELECT, EZFILE, EZLABEL, EZTEXT, EZPRESETS, EZDIR = range(14)
SKINTYPE, EYETYPE, NAILTYPE, MOUTHTYPE, TEETHTYPE, OTHERTYPE = range(6)
METHOD, SHADERTYPE = range(2)

# This class must exist - EZSkin instantiates it for building the shaders
class EZBuilder(object):

    def __init__(self, api):

        self.api = api                            # A reference to an EZAPI object
        self.parameters = None                     # The method to interrogate the value of UI controls (this will be set by EZSkin)

        self.has_controls = True                # True if you supply custom controls
        self.unique_name = '02_EZ_CycleSkin'    # Sorry - no naming convention yet

        # This dictionary is used to associate an EZSkin shader with a method in your classs
        # In this case, we only have one method so all are set to invoke self.MakePBR
        # The second element of the tuple determines whether or not it should be processed,
        # As selected by the user on the main EZSkin UI
        self.shaders = {
            "cornea" : (self.CycleCornea, EYETYPE),
            "eyelash" : (self.CycleLash, OTHERTYPE),
            "eyetrans" : (self.CycleTrans, EYETYPE),
            "gums" : (self.Cyclemouth, MOUTHTYPE),
            "hair" : (self.CycleHair, OTHERTYPE),
            "ignore" : (None, SKINTYPE),
            "innermouth" : (self.Cyclemouth, MOUTHTYPE),
            "invisible" : (self.CycleTrans, OTHERTYPE),
            "iris" : (self.CycleIris, EYETYPE),
            "lacrimal" : (self.Cyclemouth, EYETYPE),
            "lips" : (self.CycleLips, SKINTYPE),
            "nails" : (self.CycleNail, NAILTYPE),
            "pupil" : (self.CyclePupil, EYETYPE),
            "sclera" : (self.CycleSclera, EYETYPE),
            "skin" : (self.CycleSkin, SKINTYPE),
            "specular" : (None, SKINTYPE),
            "tear" : (self.CycleTrans, EYETYPE),
            "teeth" : (self.CycleTeeth, TEETHTYPE),
            "tongue" : (self.Cyclemouth, MOUTHTYPE),
            "suntan" : (self.CycleSuntan, SKINTYPE),
                        "face" : (self.CycleFace, SKINTYPE),
            }

    def _PrepareRoot(self, mat, p):
        # Used to prepare the material room for a new Superfly shader
        # In this case it will return a PoserSurface root node
        # User preferences will be taken into account (use existing root or make a new one)
        # All existing roots are managed for you

        # Get original CyclesSurface root node
        n000 = mat.roots['superfly']
        if n000:
            if n000.Type() != poser.kNodeTypeCodeCYCLESSURFACE:
                n000 = None

        # Make new root
        if p('make_root', 'settings') or n000 is None:
            n000 = mat.MakeRoot(renderer='superfly', roottype=poser.kNodeTypeCodeCYCLESSURFACE)
            mat.SetPSRoot(n000)

        return n000

    def BuilderName(self):
        # REQUIRED: returns the name of the 'shader engine' to be displayed to the user
        return 'EZ_CycleSkin (SF)'

    def BuildShader(self, material, data, mat):

        # Parse the data
        shader = data[0]
        scattergroup = int(data[1])

        # Don't waste time if we don't have to do anything
        if shader == 'ignore': return 'ignored'

        # Stop if no PoserSurface is found
        if mat.Root() is None: return 'skipped - No Firefly PoserSurface found'

        # Configure imagemap nodes
        p = self.parameters
        if p('gc'): mat.GC_AllMaps()
        if p('filter'): mat.ApplyTextureFilters(p('filtering') + 1)

        # ... and build the shader
        if shader in self.Shaders():

            # Get the method which will build the shader
            method = self.shaders[shader][METHOD]

            # Check to see if we need to use a built-in shader
            if method is None and self.api.Build() >= 68:
                builder = self.api.GetBuilder(self.Renderer())
                if builder:
                    return builder.BuildShader(material, data, mat)
                else:
                    return 'failed'

            else:
                # Build the shader
                if method is not None:
                    method(mat, p)
                    mat.RemoveOrphans()
                    mat.UpdatePreview()

                return 'processed'


    def Controls(self):
        # If has_controls = True then this method must be present and return a tuple of tuples
        # (name, label, kind, minmax, default, increment=None, tooltip='', category=None)
        # See controls.txt for examples

        # Define the controls
        # (self, name, label, kind, minmax, default, increment=None, tooltip='', category=None)
        data = (
            
            ('02_highgloss', 'HighGloss', EZFLOAT, (0.0001, 1.0), 0.17, 0.01, 'High gloss value', 'Main'),
            ('02_hgcolor', 'HighGloss Colour', EZCOLOURP, (None), (0.85, 0.85, 0.85), None, 'HighGloss Colour', 'Main'),
            ('02_lowgloss', 'LowGloss', EZFLOAT, (0.0001, 1.0), 0.45, 0.01, 'Low gloss value', 'Main'),
            ('02_lgcolor', 'LowGloss Colour', EZCOLOURP, (None), (0.9, 0.9, 0.9), None, 'LowGloss Colour', 'Main'),
            ('02_lipgloss', 'LipGloss', EZFLOAT, (0.0001, 1.0), 0.15, 0.01, 'Lip gloss value', 'Main'),
            ('02_lpcolor', 'LipGloss Colour', EZCOLOURP, (None), (1.0, 1.0, 1.0), None, 'LipGloss Colour', 'Main'),
            ('02_sssdepth', 'SSS_Depth', EZFLOAT, (0.1,100.0), 2.5, 0.1, 'SubSurface Scattering depth(in mm.)', 'Main'),
            ('02_gradbump', 'Gradient Bump map', EZBOOL, (None), 0, None, 'Gradient Bump map', 'Main'),
            ('02_bumpheight', 'BumpHeight', EZFLOAT, (0.01,100.0), 1.0, 1.0, 'Bump height(in mm.)', 'Main'),
            ('02_procbump', 'Procedural Bump', EZBOOL, (None), 0, None, 'Use procedural bump', 'Main'),
            ('02_procbumpheight', 'ProcBumpHeight', EZFLOAT, (0.1,100.0), 0.3, 1.0, 'Procedural Bump height(in mm.)', 'Main'),
            ('02_suntanmask', 'Suntan Mask', EZFILE, (None), "Suntanmask_T2.jpg", None, 'Mask for Suntan', 'Main'),
            ('02_suntanintensity', 'SuntanIntensity', EZFLOAT, (0.75, 0.95), 0.81, 0.05, 'value between 0.75 and 0.95', 'Main'),
            ('02_hairshinecolor', 'Hair Shine Color', EZCOLOURP, (None), (0.75,0.75,0.75), None, 'Hair Shine ', 'Main'),
            ('02_hairroughness', 'Hair Roughness', EZFLOAT, (0.05, 0.7), 0.2, 0.01, 'Hair Roughness', 'Main'),
            ('02_wetskin', 'Wet Skin', EZBOOL, (None), 0, None, 'Make wet skin', 'Main'),
            ('02_foam', 'Foamy Skin', EZBOOL, (None), 0, None, 'Apply foam', 'Main'),
            ('02_effects', 'Special Effects', EZCHOICE, ('Nothing','Grease', 'Yellow Custard', 'Mud'), 0, None, 'Integer', 'Main'),
            ('02_custardcolor', 'Custard Colour', EZCOLOURP, (None), (1.0, 1.0, 0.0), None, 'Custard color', 'Main'),
            ('02_mudcolor', 'Mud Colour', EZCOLOURP, (None), (0.5450,0.4980,0.40392), None, 'Mud color', 'Main'),
            ('02_glossyeffect', 'Effects Glossyness', EZFLOAT, (0.1, 0.5), 0.2, 0.1, 'Glossyness effects', 'Main'),
            ('02_nowetface', 'No Wet Face', EZBOOL, (None), 1, None, 'No wet face', 'Main'),
            ('02_nofoamyface', 'No Foamy Face', EZBOOL, (None), 1, None, 'No foam on face', 'Main'),
            )

        return data

    def Nix(self, mat, p):
        pass

    def Shaders(self):
        # Used internally by EZSkin
        return list(self.shaders.keys())

    def ShaderType(self, shader):
        # Used internally by EZSkin
        return self.shaders.get(shader, (None, 0))[SHADERTYPE]

        #Some Functions to use with the shader

    def WetSkin(self, mat, nod1, nod2):
                
        mapdict ={ 'Vector Type' : 4.0, 'Rotation' : (0.0,0.0,1.570796), 'Scale' : (3.5,1.0,3.5) }
        nw01 = mat.AddNode('ccl_MixClosure', name = 'WetClosure', loc=(200,800), collapsed=False, preview=False, )
        nw02 = mat.AddNode('ccl_Fresnel', name='Fresnel1', loc=(200,600), collapsed=False, preview=False, IOR=1.33)
        nw03 = mat.AddNode('ccl_Bump', name='WetBump', loc=(400,800), collapsed=False, preview=False, Distance=0.002,Invert  = 1 )
        nw04 = mat.AddNode('ccl_GlossyBsdf', name = 'WetGloss', loc=(400,1200), collapsed=False, preview=False, Color = (0.9,0.9,0.9), Roughness=0.02, Distribution = 3)
        nw05 = mat.AddNode('ccl_MusgraveTexture', name = 'Waterdrops', loc=(800,1200), collapsed=False, preview=False, Scale=45.0, Detail = 7, Type = 1.0,)
        nw06 = mat.AddNode('ccl_Mapping', name = 'Vertical Drops', loc=(1000,1200), collapsed=False, preview=False, inputs =mapdict)
        nw07 = mat.AddNode('ccl_TextureCoordinate', name = 'TextureCoordinates', loc=(1200,1200), collapsed=False, preview=False)
        nw08 = mat.AddNode('ccl_MixClosure', name = 'OverallClosure', loc=(450,600), collapsed=False, preview=False, Fac =0.18,)
        nw09 = mat.AddNode('ccl_GlossyBsdf', name =  'OverallGloss', loc =(900,600) , collapsed=False, preview=False, Color = (0.5,0.5,0.5), Roughness = 0.08, Distribution = 3)
            
                
        # Connect Nodes
        mat.ConnectNodes(nw01, nod1, 'Surface')
        mat.ConnectNodes(nod2, nw08, 'Closure1')            
        mat.ConnectNodes(nw08, nw01, 'Closure1')
        mat.ConnectNodes(nw04, nw01, 'Closure2')
        mat.ConnectNodes(nw02, nw01, 'Fac')
        mat.ConnectNodes(nw09, nw08, 'Closure2')
        mat.ConnectNodes(nw03, nw09, 'Normal')
        mat.ConnectNodes(nw03, nw04, 'Normal')
        mat.ConnectNodes(nw05, nw03, 'Height')
        mat.ConnectNodes(nw06, nw05, 'Vector')
        mat.ConnectNodes(nw07, nw06, 'Vector')
        return nw01

    def FoamySkin(self, mat, p, nod1, nod2, nod3):
        FSC_Dict = {'Color1' :(0.98,0.98,0.47), 'Color2' : (0.76,1.0,0.47), 'Color3':(0.976,0.619,0.99), 'Color4' :(1.0,0.976,0.929)}
        FS_Dict = {'x_Index' : 4.0, 'y_Index': 4.0, 'z_Index': 4.0, 'Octaves' : 5.0, 'Bottom': 0.4, 'Bias': 0.6, 'Gain' : 0.5, 'Offset_X': 0.5, 'Offset_Y': 0.5, 'Offset_Z': 0.5}
        TB_Dict = {'x_Index' : 0.5, 'y_Index': 0.5, 'z_Index': 0.5, 'Octaves' : 4.0, 'Bottom': 0.0, 'Bias': 0.5, 'Gain' : 0.5, 'Noise_Type' : 2,'Offset_X': 0.5, 'Offset_Y': 0.5, 'Offset_Z': 0.5}
        nf01 = mat.AddNode('ccl_MixClosure', name = 'FoamClosure', loc=(20,800), collapsed=False, preview=False,)
        nf02 = mat.AddNode('ccl_MixClosure', name = 'FoamClosure1', loc=(20,1000), collapsed=False, preview=False,)
        nf03 = mat.AddNode('ccl_MixClosure', name = 'FoamClosure2', loc=(200,1000), collapsed=False, preview=False,)
        nf04 = mat.AddNode('ccl_Fresnel', name='Fresnel foam', loc=(200,800), collapsed=False, preview=False, IOR=1.33)
        nf05 = mat.AddNode('ccl_SubsurfaceScattering', name='FoamSSS', loc=(400,800), collapsed=False, preview=False, Scale=0.75, Falloff=1,)
        nf06 = mat.AddNode('ccl_DiffuseBsdf', name = 'White foam', loc=(400,1000),  collapsed=False, preview=False, Color=(1,1,1), Roughness=0.2,)
        nf07 = mat.AddNode('ccl_GlossyBsdf', name = 'FoamGloss', loc=(400,1200), collapsed=False, preview=False, Roughness=0.03, Distribution = 3)
        nf08 = mat.AddNode('cellular', name='Final Foam Shape', loc=(600,800), collapsed=False, preview=False, Color1=(1.0,1.0,1.0), Color2=(0.98,0.98,0.98), ScaleX=0.1,ScaleY=0.2,ScaleZ=0.1, cellMode=3, Jitter=0.939, Turbulence=0.862, Noise_Type=2,)
        nf09 = mat.AddNode('hsv', name='Scattercolor', loc=(600,1200), collapsed=False, preview=False, Saturation =0.5,)
        nf10 = mat.AddNode('colorramp', name='FoamShineColor', loc=(600,1400), collapsed=False, preview=False, inputs=FSC_Dict)
        nf11 = mat.AddNode('ccl_Bump', name='FoamBump', loc=(600,1600), collapsed=False, preview=False, Distance=0.03)
        nf12 = mat.AddNode('marble', name='FoamPattern', loc=(800,800),collapsed=False, preview=False, Base_Color=(1,1,1), Veins=(0.77,0.776,0.776), Scale=1.59, Turbulence=2.264, Noise_Type=2,)
        nf13 = mat.AddNode('fractal_sum', name='FoamShape', loc=(800,1000),collapsed=False, preview=False, inputs=FS_Dict)
        nf14 = mat.AddNode('turbulence', name= 'Bubbles', loc=(800,1400),collapsed=False, preview=False, inputs=TB_Dict)

        # Connect Nodes
        mat.ConnectNodes(nf01, nod1, 'Surface')
        if p('02_wetskin') is True:
                mat.ConnectNodes(nod3, nf01, 'Closure1')
        else:
                mat.ConnectNodes(nod2, nf01, 'Closure1')                       
        mat.ConnectNodes(nf02, nf01, 'Closure2')
        mat.ConnectNodes(nf08, nf01, 'Fac')
        mat.ConnectNodes(nf03, nf02, 'Closure1')
        mat.ConnectNodes(nf07, nf02, 'Closure2')
        mat.ConnectNodes(nf04, nf02, 'Fac')
        mat.ConnectNodes(nf08, nf03, 'Fac')
        mat.ConnectNodes(nf05, nf03, 'Closure1')
        mat.ConnectNodes(nf06, nf03, 'Closure2')
        mat.ConnectNodes(nf08, nf05, 'Color')
        mat.ConnectNodes(nf11, nf05, 'Normal')
        mat.ConnectNodes(nf09, nf05, 'Radius')
        mat.ConnectNodes(nf11, nf06, 'Normal')
        mat.ConnectNodes(nf10, nf07, 'Color')
        mat.ConnectNodes(nf11, nf07, 'Normal')
        mat.ConnectNodes(nf13, nf08, 'Color1')
        mat.ConnectNodes(nf13, nf08, 'Color2')
        mat.ConnectNodes(nf12, nf08, 'ScaleX')
        mat.ConnectNodes(nf12, nf08, 'ScaleY')
        mat.ConnectNodes(nf12, nf08, 'ScaleZ')
        mat.ConnectNodes(nf10, nf09, 'Color')
        mat.ConnectNodes(nf08, nf11, 'Height')
        mat.ConnectNodes(nf14, nf10, 'Input')


    def GreasySkin(self, mat, p, nod1, nod2):
        clodict = {'Sky_Color' : (1.0,1.0,1.0), 'Cloud_Color' : (0.0,0.0,0.0), 'Scale': 3, 'Complexity' : 6.0, 'Bottom': 0.2, 'Bias': 0.5, 'Gain': 0.6}
        ng01=mat.AddNode('ccl_MixClosure', name = 'GreaseClosure1', loc=(200,1200),collapsed=False, preview=False,)
        ng02=mat.AddNode('ccl_MixClosure', name = 'GreaseClosure2', loc=(450,1300),collapsed=False, preview=False,)
        ng03=mat.AddNode('ccl_MixClosure', name = 'GreaseClosure1', loc=(700,1400),collapsed=False, preview=False, Fac=p('02_glossyeffect'),)
        ng04=mat.AddNode('ccl_Fresnel', name = 'GreaseFresnel', loc=(700,1200),collapsed=False, preview=False, IOR=1.5,)
        ng05=mat.AddNode('ccl_GlossyBsdf', name = 'HighGreaseGlow', loc=(700,1600),collapsed=False, preview=False, Color=(0.776,0.776,0.776), Roughness=0.05,)
        ng06=mat.AddNode('ccl_GlossyBsdf', name = 'LowGreaseGlow', loc=(950,1500),collapsed=False, preview=False, Color=(0.4,0.4,0.4), Roughness=0.05,)
        ng07=mat.AddNode('ccl_DiffuseBsdf', name = 'GreaseColor', loc=(950,1300),collapsed=False, preview=False, Color=(0.2235,0.1568,0), Roughness=0.3,)
        ng08=mat.AddNode('math_functions', name = 'ReverseClouds',loc=(950,1100),collapsed=False, preview=False, Math_Argument = 2, Value_1 = 1.0, Value_2=1.0,)
        ng09=mat.AddNode('clouds',name = 'GreasePattern', loc = (1200,1100),collapsed=False, preview=False, inputs = clodict,)

        # Connect Nodes

        mat.ConnectNodes(ng01, nod1, 'Surface')
        mat.ConnectNodes(nod2, ng01, 'Closure1')
        mat.ConnectNodes(ng02, ng01, 'Closure2')
        mat.ConnectNodes(ng08, ng01, 'Fac')
        mat.ConnectNodes(ng04, ng02, 'Fac')
        mat.ConnectNodes(ng03, ng02, 'Closure1')
        mat.ConnectNodes(ng05, ng02, 'Closure2')
        mat.ConnectNodes(ng07, ng03, 'Closure1')
        mat.ConnectNodes(ng06, ng03, 'Closure2')
        mat.ConnectNodes(ng09, ng08, 'Value_2')

                        
    def YellowCustard(self, mat, p, nod1, nod2, haircheck):

        fracdict = {'Octaves': 4.0 , 'Bottom': 0.2 , 'Bias' : 0.6 , 'Gain': 0.6}
        ny01=mat.AddNode('ccl_MixClosure', name = 'YecClosure1', loc=(200,1200),collapsed=False, preview=False,)
        ny02=mat.AddNode('ccl_MixClosure', name = 'YecClosure2', loc=(450,1200),collapsed=False, preview=False, Fac=p('02_glossyeffect'),)
#        ny03=mat.AddNode('ccl_Fresnel', name = 'YecFresnel', loc=(700,800),collapsed=False, preview=False, IOR= 1.35,)
        ny04=mat.AddNode('ccl_SubsurfaceScattering', name = 'YecSSS', loc=(700,1200),collapsed=False, preview=False, Color = p('02_custardcolor'), Scale = 0.006,Radius = (1.0,1.0,0.5),)
        ny05=mat.AddNode('ccl_GlossyBsdf', name = 'YecGlossy', loc=(700,1600),collapsed=False, preview=False, Color = (0.77,0.77,0.77), Roughness = 0.05,)
        ny06=mat.AddNode('ccl_Bump', name = 'YecBump', loc = (900,1200),collapsed=False, preview=False, Distance = 0.01,)
        ny07=mat.AddNode('fractal_sum', name = 'Yecpattern', loc=(1150,1200),collapsed=False, preview=False, inputs = fracdict,)

        # ConnectNodes

        if haircheck is False :
                mat.ConnectNodes(ny01, nod1, 'Surface')
        else:
                mat.ConnectNodes(ny01, nod1, 'Closure2')
        mat.ConnectNodes(nod2, ny01, 'Closure1')
        mat.ConnectNodes(ny02, ny01, 'Closure2')
        mat.ConnectNodes(ny07, ny01, 'Fac')
#        mat.ConnectNodes(ny03, ny02, 'Fac')
        mat.ConnectNodes(ny05, ny02, 'Closure2')
        mat.ConnectNodes(ny04, ny02, 'Closure1')
#         mat.ConnectNodes(ny06, ny03, 'Normal')
        mat.ConnectNodes(ny06, ny04, 'Normal')
        mat.ConnectNodes(ny06, ny05, 'Normal')
        mat.ConnectNodes(ny07, ny06, 'Height')
                             



    def Mud(self, mat, p, nod1, nod2, haircheck):

        mapdict ={ 'Vector Type' : 4.0, 'Rotation' : (0.0,0.0,1.570796), 'Scale' : (2.5,1.0,2.5) }
        fracdict = {'Octaves': 5.0 , 'Bottom': 0.1 , 'Bias' : 0.8 , 'Gain': 0.6}
        nm01=mat.AddNode('ccl_MixClosure', name = 'MudClosure1', loc=(200,1200),collapsed=False, preview=False,)
        nm02=mat.AddNode('ccl_MixClosure', name = 'MudClosure2', loc=(450,1200),collapsed=False, preview=False, Fac=p('02_glossyeffect'))
        nm04=mat.AddNode('ccl_DiffuseBsdf', name = 'MudColor', loc=(700,1200),collapsed=False, preview=False, Color = p('02_mudcolor'),Roughness = 0.3,)
        nm05=mat.AddNode('ccl_GlossyBsdf', name = 'MudGlossy', loc=(700,1400),collapsed=False, preview=False, Color = (0.5,0.5,0.5), Roughness = 0.05,)
        nm06=mat.AddNode('ccl_Bump', name = 'MudBump', loc=(900,1200),collapsed=False, preview=False, Distance = 0.002,)
        nm07=mat.AddNode('fractal_sum', name = 'Mudpattern', loc=(1150,1200),collapsed=False, preview=False, inputs = fracdict,)
        nm08 = mat.AddNode('ccl_MusgraveTexture', name = 'Mudstructure', loc=(800,1200), collapsed=False, preview=False, Scale=40.0, Detail = 2, Type = 1.0,)
        nm09 = mat.AddNode('ccl_Mapping', name = 'Vertical Drops', loc=(1000,1200), collapsed=False, preview=False, inputs =mapdict)
        nm10 = mat.AddNode('ccl_TextureCoordinate', name = 'TextureCoordinates', loc=(1200,1200), collapsed=False, preview=False)

        # ConnectNodes

        if haircheck is False :
                mat.ConnectNodes(nm01, nod1, 'Surface')
        else:
                mat.ConnectNodes(nm01, nod1, 'Closure2')
        mat.ConnectNodes(nod2, nm01, 'Closure1')
        mat.ConnectNodes(nm02, nm01, 'Closure2')
        mat.ConnectNodes(nm07, nm01, 'Fac')
        mat.ConnectNodes(nm04, nm02, 'Closure1')
        mat.ConnectNodes(nm05, nm02, 'Closure2')
        mat.ConnectNodes(nm06, nm04, 'Normal')
        mat.ConnectNodes(nm06, nm05, 'Normal')
        mat.ConnectNodes(nm08, nm06, 'Height')
        mat.ConnectNodes(nm09, nm08, 'Vector')
        mat.ConnectNodes(nm10, nm09, 'Vector')
                                 
                
                                        
    

    # Our First shader building method
    def CycleSkin(self, mat, p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)



        #Convert Input parameters
        BmpHgt=p('02_bumpheight')/1000.0
        SSSDpt=p('02_sssdepth')/1000.0
        PrcBmpHgt=p('02_procbumpheight')/1000.0

        # Prepare Node
        n001 = mat.AddNode('ccl_MixClosure', name='MixClosure1', loc=(300,20), collapsed=False, preview=False, )
        n002 = mat.AddNode('ccl_MixClosure', name='MixClosure2', loc=(600,200), collapsed=False, preview=False, Fac=0.273)
        n003 = mat.AddNode('ccl_Fresnel', name='Fresnel1', loc=(600,20), collapsed=False, preview=False, IOR=1.45)
        n004 = mat.AddNode('ccl_SubsurfaceScattering', name='SSS', loc=(900,20), collapsed=False, preview=False, Scale=SSSDpt, Radius=(3.67,1.76,0.6), Falloff=1,)
        n005 = mat.AddNode('ccl_GlossyBsdf', name='Lowgloss', loc=(900,200), collapsed=False, preview=False, Roughness=p('02_lowgloss'), )
        n007 = mat.AddNode('ccl_HSV', name='LowglossColor', loc=(1200,200), collapsed=False, preview=False, Color=p('02_lgcolor'), )
        n006 = mat.AddNode('ccl_GlossyBsdf', name='Highgloss', loc=(900,400), collapsed=False, preview=False, Roughness=p('02_highgloss'), )
        n008 = mat.AddNode('ccl_HSV', name='HighglossColor', loc=(1200,400), collapsed=False, preview=False, Color=p('02_hgcolor'), )
        n009 = mat.AddNode('ccl_SeparateRGB', name='Redfilter', loc=(1500,200), collapsed=False, preview=False )
        n010 = mat.AddNode('ccl_Bump', name='Bump2Normal', loc=(1500,400), collapsed=False, preview=False, Distance=BmpHgt,)
        n011 = mat.GetMap('color')
        n012 = mat.GetMap('bump')
        n013 = mat.GetMap('normal')


        if n011 is None :
            n011=mat.AddNode('simple_color', name='Diffuse Color', loc=(700,20), collapsed=False, preview=False, )
        if n012 is None and n013 is None or p('02_procbump') is True:
            n014=mat.AddNode('ccl_VoronoiTexture', name='Pores', loc=(1800,20), collapsed=False, preview=False, Scale=1050, Coloring=2,)
            n015=mat.AddNode('ccl_Invert', name='Inverted Redfilter', loc=(1800,200), collapsed=False, preview=False,)
            n016=mat.AddNode('ccl_Bump', name='Make pores', loc=(1800,400), collapsed=False, preview=False, Distance=PrcBmpHgt, )
            n017=mat.AddNode('ccl_Math', name='Multiply', loc =(1800, 600), collapsed=False, preview=False, Type=20, )
        

        # Connect nodes
        mat.ConnectNodes(n001, n000, 'Surface')
        mat.ConnectNodes(n002, n001, 'Closure1')
        mat.ConnectNodes(n003, n001, 'Fac')
        mat.ConnectNodes(n004, n002, 'Closure1')
        mat.ConnectNodes(n005, n002, 'Closure2')
        mat.ConnectNodes(n006, n001, 'Closure2')
        mat.ConnectNodes(n011, n004, 'Color')
        mat.ConnectNodes(n007, n005, 'Color')
        mat.ConnectNodes(n008, n006, 'Color')
        mat.ConnectNodes(n009, n007, 'Value')
        mat.ConnectNodes(n009, n008, 'Value')
        mat.ConnectNodes(n009, n004, 'Texture Blur')        
        mat.ConnectNodes(n011, n009, 'Image')
        mat.ConnectNodes(n010, n003, 'Normal')
        mat.ConnectNodes(n010, n004, 'Normal')
        mat.ConnectNodes(n010, n005, 'Normal')
        mat.ConnectNodes(n010, n006, 'Normal')
        if n012 is None and n013 is None or p('02_procbump') is True:
                mat.ConnectNodes(n009, n015, 'Color')
                mat.ConnectNodes(n015, n010, 'Height')
                mat.ConnectNodes(n009, n017, 'Value1')
                mat.ConnectNodes(n014, n017, 'Value2')
                mat.ConnectNodes(n017, n016, 'Height')
                mat.ConnectNodes(n016, n010, 'Normal')
        else:
                if n012 :
                        mat.ConnectNodes(n012, n010, 'Height')
                elif n013 :
                        if p('02_gradbump') is True :
                                mat.ConnectNodes(n013, n010, 'Height')
                        else: 
                                mat.ConnectNodes(n013, n010, 'Normal')

        # in case of displacement file

        n017 = mat.GetMap('displacement')        
        if n017 :
                dispval = mat.PSValue('Displacement')
                n018 = mat.AddNode('ccl_HSV', name = 'Disp_Control', loc = (250,400), collapsed=False, preview=False, Value = dispval,)
                n019 = mat.AddNode('math_functions', name = 'Reduce_D_Map)', loc = (500,400), collapsed=False, preview=False, Math_Argument=2.0, Value_2=0.5,)
                mat.ConnectNodes(n018, n000, 'Displacement')
                mat.ConnectNodes(n019, n018, 'Color')
                mat.ConnectNodes(n017, n019, 'Value_1')
                        

        # Wetaddon
        nw01 = None
        
        if p('02_wetskin') is True:
                        nw01=self.WetSkin(mat, n000, n001)


        # Foamaddon
        

        if p('02_foam') is True:
            self.FoamySkin(mat, p, n000, n001, nw01)



        #Greasy

        if p('02_effects') == 1:
                self.GreasySkin(mat, p, n000, n001)

        # Yellow Custard Cover

        if p('02_effects') == 2:
                self.YellowCustard(mat, p, n000, n001, False)

        # Muddy

        if p('02_effects') == 3:
                self.Mud(mat, p, n000, n001,False)


                        
                        
            

                        
    # Second Shader
    def CycleLips(self, mat ,p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        #Convert Input parameters
        BmpHgt=p('02_bumpheight')/1000.0
        SSSDpt=p('02_sssdepth')/1000.0
        PrcBmpHgt=p('02_procbumpheight')/1000.0
        
        # Prepare Node
        n001 = mat.AddNode('ccl_MixClosure', name='MixClosure1', loc=(300,20), collapsed=False, preview=False, )
        n002 = mat.AddNode('ccl_MixClosure', name='MixClosure2', loc=(600,200), collapsed=False, preview=False, Fac=0.273)
        n003 = mat.AddNode('ccl_Fresnel', name='Fresnel1', loc=(600,20), collapsed=False, preview=False, IOR=1.5)
        n004 = mat.AddNode('ccl_SubsurfaceScattering', name='SSS', loc=(900,20), collapsed=False, preview=False, Scale=SSSDpt, Radius=(3.67,1.76,0.6), Falloff=1,)
        n005 = mat.AddNode('ccl_GlossyBsdf', name='Lowgloss', loc=(900,200), collapsed=False, preview=False, Roughness=p('02_lipgloss'), )
        n007 = mat.AddNode('ccl_HSV', name='LipglossColor', loc=(1200,200), collapsed=False, preview=False, Color=p('02_lpcolor'), )
        n006 = mat.AddNode('ccl_GlossyBsdf', name='Highgloss', loc=(900,400), collapsed=False, preview=False, Roughness=p('02_highgloss'), )
        n008 = mat.AddNode('ccl_HSV', name='HighglossColor', loc=(1200,400), collapsed=False, preview=False, Color=p('02_hgcolor'), )
        n009 = mat.AddNode('ccl_SeparateRGB', name='Redfilter', loc=(1500,200), collapsed=False, preview=False )
        n010 = mat.AddNode('ccl_Bump', name='Bump2Normal', loc=(1500,400), collapsed=False, preview=False, Distance=BmpHgt,)
        n011 = mat.GetMap('color')
        n012 = mat.GetMap('bump')
        n017 = mat.GetMap('normal')


        if n011 is None:
            n011=mat.AddNode('simple_color', name='Diffuse Color', loc=(700,20), collapsed=False, preview=False, )
        if n012 is None and n017 is None or p('02_procbump') is True:
            n013=mat.AddNode('ccl_VoronoiTexture', name='Pores', loc=(1800,20), collapsed=False, preview=False, Scale=1050, Coloring=2,)
            n014=mat.AddNode('ccl_Invert', name='Inverted Redfilter', loc=(1800,200), collapsed=False, preview=False,)
            n015=mat.AddNode('ccl_Bump', name='Make pores', loc=(1800,400), collapsed=False, preview=False, Distance=PrcBmpHgt, )
            n016=mat.AddNode('ccl_Math', name='Multiply', loc =(1800, 600), collapsed=False, preview=False, Type=20, )
        

        # Connect nodes
        mat.ConnectNodes(n001, n000, 'Surface')
        mat.ConnectNodes(n002, n001, 'Closure1')
        mat.ConnectNodes(n003, n001, 'Fac')
        mat.ConnectNodes(n004, n002, 'Closure1')
        mat.ConnectNodes(n005, n002, 'Closure2')
        mat.ConnectNodes(n006, n001, 'Closure2')
        mat.ConnectNodes(n011, n004, 'Color')
        mat.ConnectNodes(n007, n005, 'Color')
        mat.ConnectNodes(n008, n006, 'Color')
        mat.ConnectNodes(n009, n007, 'Value')
        mat.ConnectNodes(n009, n008, 'Value')
        mat.ConnectNodes(n009, n004, 'Texture Blur')    
        mat.ConnectNodes(n011, n009, 'Image')
        mat.ConnectNodes(n010, n003, 'Normal')
        mat.ConnectNodes(n010, n004, 'Normal')
        mat.ConnectNodes(n010, n005, 'Normal')
        mat.ConnectNodes(n010, n006, 'Normal')
        if n012 is None and n017 is None or p('02_procbump') is True:
                mat.ConnectNodes(n009, n010, 'Height')
                mat.ConnectNodes(n009, n016, 'Value1')
                mat.ConnectNodes(n013, n016, 'Value2')
                mat.ConnectNodes(n016, n015, 'Height')
                mat.ConnectNodes(n015, n010, 'Normal')
        else:
                if n012 :
                        mat.ConnectNodes(n012, n010, 'Height')
                elif n017 :
                        if p('02_gradbump') is True :
                                mat.ConnectNodes(n017, n010, 'Height')
                        else: 
                                mat.ConnectNodes(n017, n010, 'Normal')

        # Wetaddon
        nw01 = None
        
        if p('02_wetskin') is True:
            nw01=self.WetSkin(mat, n000, n001)


        # Foamaddon
        

        if p('02_foam') is True:
            self.FoamySkin(mat, p, n000, n001, nw01)



        #Greasy

        if p('02_effects') == 1:
                self.GreasySkin(mat, p, n000, n001)

        # Yellow Custard Cover

        if p('02_effects') == 2:
                self.YellowCustard(mat, p, n000, n001, False)

        # Muddy

        if p('02_effects') == 3:
                self.Mud(mat, p, n000, n001, False)

                                         
    # Our Third shader building method
    def Cyclemouth(self, mat, p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        #Convert Input parameters
        BmpHgt=p('02_bumpheight')/1000.0
        SSSDpt=p('02_sssdepth')/1000.0
        PrcBmpHgt=p('02_procbumpheight')/1000.0

        # Prepare Node
        n001 = mat.AddNode('ccl_MixClosure', name='MixClosure1', loc=(300,20), collapsed=False, preview=False, )
        n002 = mat.AddNode('ccl_MixClosure', name='MixClosure2', loc=(600,200), collapsed=False, preview=False, Fac=0.273)
        n003 = mat.AddNode('ccl_Fresnel', name='Fresnel1', loc=(600,20), collapsed=False, preview=False, IOR=1.33)
        n004 = mat.AddNode('ccl_SubsurfaceScattering', name='SSS', loc=(900,20), collapsed=False, preview=False, Scale=SSSDpt, Radius=(3.67,1.76,0.6), Falloff=1,)
        n005 = mat.AddNode('ccl_GlossyBsdf', name='Lowgloss', loc=(900,200), collapsed=False, preview=False, Roughness=0.25, )
        n007 = mat.AddNode('ccl_HSV', name='LowglossColor', loc=(1200,200), collapsed=False, preview=False, Color=(0.75,0.75,0.75), )
        n006 = mat.AddNode('ccl_GlossyBsdf', name='Highgloss', loc=(900,400), collapsed=False, preview=False, Roughness=0.17, )
        n008 = mat.AddNode('ccl_HSV', name='HighglossColor', loc=(1200,400), collapsed=False, preview=False, Color=(0.94,0.94,0.94), )
        n009 = mat.AddNode('ccl_SeparateRGB', name='Redfilter', loc=(1500,200), collapsed=False, preview=False )
        n010 = mat.AddNode('ccl_Bump', name='Bump2Normal', loc=(1500,400), collapsed=False, preview=False, Distance=BmpHgt,)
        n011 = mat.GetMap('color')
        n012 = mat.GetMap('bump')
        n017 = mat.GetMap('normal')


        if n011 is None:
            n011=mat.AddNode('simple_color', name='Diffuse Color', loc=(700,20), collapsed=False, preview=False, )
        if n012 is None and n017 is None or p('02_procbump') is True:
            n013=mat.AddNode('ccl_VoronoiTexture', name='Pores', loc=(1800,20), collapsed=False, preview=False, Scale=1050, Coloring=2,)
            n014=mat.AddNode('ccl_Invert', name='Inverted Redfilter', loc=(1800,200), collapsed=False, preview=False,)
            n015=mat.AddNode('ccl_Bump', name='Make pores', loc=(1800,400), collapsed=False, preview=False, Distance=PrcBmpHgt, )
            n016=mat.AddNode('ccl_Math', name='Multiply', loc =(1800, 600), collapsed=False, preview=False, Type=20, )
        

        # Connect nodes
        mat.ConnectNodes(n001, n000, 'Surface')
        mat.ConnectNodes(n002, n001, 'Closure1')
        mat.ConnectNodes(n003, n001, 'Fac')
        mat.ConnectNodes(n004, n002, 'Closure1')
        mat.ConnectNodes(n005, n002, 'Closure2')
        mat.ConnectNodes(n006, n001, 'Closure2')
        mat.ConnectNodes(n011, n004, 'Color')
        mat.ConnectNodes(n007, n005, 'Color')
        mat.ConnectNodes(n008, n006, 'Color')
        mat.ConnectNodes(n009, n007, 'Value')
        mat.ConnectNodes(n009, n008, 'Value')
        mat.ConnectNodes(n009, n004, 'Texture Blur')        
        mat.ConnectNodes(n011, n009, 'Image')
        mat.ConnectNodes(n010, n003, 'Normal')
        mat.ConnectNodes(n010, n004, 'Normal')
        mat.ConnectNodes(n010, n005, 'Normal')
        mat.ConnectNodes(n010, n006, 'Normal')
        if n012 is None and n017 is None or p('02_procbump') is True:
                mat.ConnectNodes(n009, n014, 'Color')
                mat.ConnectNodes(n014, n010, 'Height')
                mat.ConnectNodes(n009, n016, 'Value1')
                mat.ConnectNodes(n013, n016, 'Value2')
                mat.ConnectNodes(n016, n015, 'Height')
                mat.ConnectNodes(n015, n010, 'Normal')
        else:
                if n012 :
                        mat.ConnectNodes(n012, n010, 'Height')
                elif n017:
                        if p('02_gradbump') is True :
                                mat.ConnectNodes(n017, n010, 'Height')
                        else: 
                                mat.ConnectNodes(n017, n010, 'Normal')
                                
                                         
     # Our Fourth shader building method
    def CycleLash(self, mat, p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        # Prepare Node
        n001 = mat.AddNode('ccl_MixClosure', name='MixClosure', loc=(200,200), collapsed=False, preview=False,)
        n002 = mat.AddNode('ccl_TransparentBsdf', name='Transparency', loc=(400,200), collapsed=False, preview=False, Color= (1.0,1.0,1.0),)
        n003 = mat.AddNode('ccl_DiffuseBsdf', name='Diffuse', loc=(400,400), collapsed=False, preview=False, Roughness=0.3,)
        n004 = mat.AddNode('ccl_Color', name='LashColor', loc=(800,400), collapsed=False, preview=False, Color=(0.1,0.05,0.05),)
        n005 = mat.GetMap('transparency')


        # Connect nodes
        mat.ConnectNodes(n001, n000, 'Surface')
        mat.ConnectNodes(n005, n001, 'Fac')
        mat.ConnectNodes(n002, n001, 'Closure1')
        mat.ConnectNodes(n003, n001, 'Closure2')
        mat.ConnectNodes(n004, n003, 'Color')




     # Our Fifth shader building method
    def CycleNail(self, mat, p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        # Prepare Node
        n001 = mat.AddNode('ccl_MixClosure', name='MixClosure1', loc=(200,20), collapsed=False, preview=False,)
        n002 = mat.AddNode('ccl_MixClosure', name='MixClosure2', loc=(400,400), collapsed=False, preview=False, Fac = 0.65,)
        n003 = mat.AddNode('ccl_Fresnel', name='Fresnel', loc=(400,20), collapsed=False, preview=False, IOR=1.6,)
        n004 = mat.AddNode('ccl_AnisotropicBsdf', name='Anisotropic', loc=(400,600), collapsed=False, preview=False, Roughness=0.05, Color=(0.8,0.8,0.8), Anisotropy=0.6,)
        n005 = mat.AddNode('ccl_GlossyBsdf', name='Glossy', loc=(600,200), collapsed=False, preview=False, Roughness=0.15)
        n006 = mat.AddNode('ccl_SubsurfaceScattering', name='SSS', loc=(600,400), collapsed=False, preview=False, Scale=0.005, Falloff=1,)
        n007 = mat.AddNode('ccl_Bump', name='Bumpconvert', loc=(800,200), collapsed=False, preview=False, Distance=0.0002,)
        n008 = mat.GetMap('color')
        n009 = mat.GetMap('bump')


        if n008 is None:
            n008=mat.AddNode('simple_color', name='Diffuse Color', loc=(700,20), collapsed=False, preview=False, )
            
        # Connect nodes
        mat.ConnectNodes(n001, n000, 'Surface')
        mat.ConnectNodes(n003, n001, 'Fac')
        mat.ConnectNodes(n002, n001, 'Closure1')
        mat.ConnectNodes(n004, n001, 'Closure2')
        mat.ConnectNodes(n005, n002, 'Closure1')
        mat.ConnectNodes(n006, n002, 'Closure2') 
        mat.ConnectNodes(n007, n003, 'Normal')
        mat.ConnectNodes(n007, n004, 'Normal')
        mat.ConnectNodes(n007, n005, 'Normal')
        mat.ConnectNodes(n007, n006, 'Normal')
        mat.ConnectNodes(n008, n005, 'Color')
        mat.ConnectNodes(n008, n006, 'Color')
        mat.ConnectNodes(n008, n006, 'Radius')
        if n009 is not None:
            mat.ConnectNodes(n009, n007, 'Height')

        # Wetaddon
        nw01 = None
        
        if p('02_wetskin') is True:
            nw01=self.WetSkin(mat, n000, n001)


        # Foamaddon
        if p('02_foam') is True:
            self.FoamySkin(mat, p, n000, n001, nw01)



        #Greasy

        if p('02_effects') == 1:
                self.GreasySkin(mat, p, n000, n001)

        # Yellow Custard Cover

        if p('02_effects') == 2:
                self.YellowCustard(mat, p, n000, n001, False)

        # Muddy

        if p('02_effects') == 3:
                self.Mud(mat, p, n000, n001,False)


     # Our Sixth shader building method
    def CycleTrans(self, mat, p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        # Prepare Node
        n001 = mat.AddNode('ccl_TransparentBsdf', name = 'Invisble', loc=(200,200), Color = (1.0,1.0,1.0),)

        # Connect nodes
        mat.ConnectNodes(n001, n000, 'Surface')


     # Our Seventh shader building method
    def CycleSclera(self, mat, p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        # Prepare Node
        n001 = mat.AddNode('ccl_MixClosure', name='MixClosure1', loc=(200,20), collapsed=False, preview=False,)
        n002 = mat.AddNode('ccl_MixClosure', name='MixClosure2', loc=(400,200), collapsed=False, preview=False, Fac=0.3,)
        n003 = mat.AddNode('ccl_Fresnel', name='Fresnel', loc=(400,20), collapsed=False, preview=False, IOR=1.33,)
        n004 = mat.AddNode('ccl_GlossyBsdf', name='HighGloss', loc=(400,400), collapsed=False, preview=False, Color=(1.0,1.0,1.0), Roughness=0.02, Distribution = 2,)
        n005 = mat.AddNode('ccl_SubsurfaceScattering', name='SSS', loc=(600,200), collapsed=False, preview=False, Scale=0.004, )
        n006 = mat.AddNode('ccl_GlossyBsdf', name='LowGloss', loc=(600,400), collapsed=False, preview=False, Color=(0.81,0.81,0.81), Roughness=0.02,)
        n008 = mat.AddNode('ccl_Color', name = 'Color', loc=(800,200), collapsed=False, preview=False, Color=(0.94,0.93,0.87),)
        n009 = mat.GetMap('color')
        if n009 is None:
            n009=mat.AddNode('simple_color', name='Diffuse Color', loc=(700,20), collapsed=False, preview=False,Color=(0.95,0.95,0.95) )

        # Connect nodes
        mat.ConnectNodes(n001, n000, 'Surface')
        mat.ConnectNodes(n003, n001, 'Fac')
        mat.ConnectNodes(n002, n001, 'Closure1')
        mat.ConnectNodes(n004, n001, 'Closure2')
        mat.ConnectNodes(n005, n002, 'Closure1')
        mat.ConnectNodes(n006, n002, 'Closure2')
        mat.ConnectNodes(n009, n005, 'Color')
        mat.ConnectNodes(n008, n005, 'Radius')


     # Our Eight shader building method
    def CycleIris(self, mat, p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        # Prepare Node
        n001 = mat.AddNode('ccl_SubsurfaceScattering', name = 'SSS', loc=(200,20), collapsed=False, preview=False, Scale=0.004,)
        n002 = mat.AddNode('ccl_Color', name= 'SSS_Color', loc=(400,20), collapsed=False, preview=False, Color=(0.93,1,0.91),)
        n003 = mat.GetMap('color')

        # Connect Nodes
        mat.ConnectNodes(n001, n000, 'Surface')
        mat.ConnectNodes(n003, n001, 'Color')
        mat.ConnectNodes(n002, n001, 'Radius')

     # Our Eight shader building method
    def CyclePupil(self, mat, p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        # Prepare Node
        n001 = mat.AddNode('ccl_DiffuseBsdf', name='Diffuse', loc=(200,20), collapsed=False, preview=False,)
        n002 = mat.GetMap('Color')
        if n002 is None:
            n002=mat.AddNode('simple_color', name='Diffuse Color', loc=(700,20), collapsed=False, preview=False, Color=(0,0,0))

        # Connect Nodes
        mat.ConnectNodes(n001, n000, 'Surface')
        mat.ConnectNodes (n002, n001, 'Color')


     # Our Nineth shader building method
    def CycleCornea(self, mat, p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        # Prepare Node
        n001 = mat.AddNode('ccl_LightPath', name='LightPath', loc=(580, 22), collapsed=False, preview=False, )
        n002 = mat.AddNode('ccl_MixClosure', name='MixClosure', loc=(300, 226), collapsed=False, preview=False, Fac=0.500, Closure1=(0.000, 0.000, 0.000), Closure2=(0.000, 0.000, 0.000), )
        n003 = mat.AddNode('ccl_GlassBsdf', name='GlassBsdf', loc=(580, 282), collapsed=False, preview=False, Color=(0.9500, 0.9500, 0.9500), Normal=(0.000, 0.000, 0.000), Roughness=0.000, IOR=1.33, Distribution=1.000, )
        n004 = mat.AddNode('ccl_TransparentBsdf', name='TransparentBsdf', loc=(580, 415), collapsed=False, preview=False, Color=(1.000, 1.000, 1.000), )

        # Connect nodes
        mat.ConnectNodes(n002, n000, 'Surface')
        mat.ConnectNodes(n001, n002, 'Fac', output="Is Shadow Ray")
        mat.ConnectNodes(n003, n002, 'Closure1')
        mat.ConnectNodes(n004, n002, 'Closure2')

     # Our tenth shader building method
    def CycleTeeth(self, mat, p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        # Prepare Node
        n001 = mat.AddNode('ccl_MixClosure', name='MixClosure1', loc=(200,20), collapsed=False, preview=False,)
        n002 = mat.AddNode('ccl_MixClosure', name='MixClosure2', loc=(400,200), collapsed=False, preview=False, Fac=0.3,)
        n003 = mat.AddNode('ccl_Fresnel', name='Fresnel', loc=(400,20), collapsed=False, preview=False, IOR=1.33,)
        n004 = mat.AddNode('ccl_GlossyBsdf', name='HighGloss', loc=(400,400), collapsed=False, preview=False, Color=(1.0,1.0,1.0), Roughness=0.02, Distribution = 2,)
        n005 = mat.AddNode('ccl_SubsurfaceScattering', name='SSS', loc=(600,200), collapsed=False, preview=False, Scale=0.008, )
        n006 = mat.AddNode('ccl_GlossyBsdf', name='LowGloss', loc=(600,400), collapsed=False, preview=False, Color=(0.81,0.81,0.81), Roughness=0.3,)
        n008 = mat.AddNode('ccl_Color', name = 'Color', loc=(800,200), collapsed=False, preview=False, Color=(0.84,1.0,0.88),)
        n009 = mat.GetMap('color')
        if n009 is None:
            n009=mat.AddNode('simple_color', name='Diffuse Color', loc=(700,20), collapsed=False, preview=False,Color=(1,0.98,0.93) )

        # Connect nodes
        mat.ConnectNodes(n001, n000, 'Surface')
        mat.ConnectNodes(n003, n001, 'Fac')
        mat.ConnectNodes(n002, n001, 'Closure1')
        mat.ConnectNodes(n004, n001, 'Closure2')
        mat.ConnectNodes(n005, n002, 'Closure1')
        mat.ConnectNodes(n006, n002, 'Closure2')
        mat.ConnectNodes(n009, n005, 'Color')
        mat.ConnectNodes(n008, n005, 'Radius')

     # Our eleventh shader building method
    def CycleSuntan(self, mat, p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        #Convert Input parameters
        BmpHgt=p('02_bumpheight')/1000.0
        SSSDpt=p('02_sssdepth')/1000.0
        PrcBmpHgt=p('02_procbumpheight')/1000.0

        # Make a dictionary for the mix node input
        sti=p('02_suntanintensity')
        mydictionary = {'Blend Type': 18, 'Color2': (sti,sti,sti)}
                

        # Prepare Node
        n001 = mat.AddNode('ccl_MixClosure', name='MixClosure1', loc=(300,20), collapsed=False, preview=False, )
        n002 = mat.AddNode('ccl_MixClosure', name='MixClosure2', loc=(600,200), collapsed=False, preview=False, Fac=0.273)
        n003 = mat.AddNode('ccl_Fresnel', name='Fresnel1', loc=(600,20), collapsed=False, preview=False, IOR=1.45)
        n004 = mat.AddNode('ccl_SubsurfaceScattering', name='SSS', loc=(900,20), collapsed=False, preview=False, Scale=SSSDpt, Radius=(3.67,1.76,0.6), Falloff=1,)
        n005 = mat.AddNode('ccl_GlossyBsdf', name='Lowgloss', loc=(900,200), collapsed=False, preview=False, Roughness=p('02_lowgloss'), )
        n007 = mat.AddNode('ccl_HSV', name='LowglossColor', loc=(1200,200), collapsed=False, preview=False, Color=p('02_lgcolor'), )
        n006 = mat.AddNode('ccl_GlossyBsdf', name='Highgloss', loc=(900,400), collapsed=False, preview=False, Roughness=p('02_highgloss'), )
        n008 = mat.AddNode('ccl_HSV', name='HighglossColor', loc=(1200,400), collapsed=False, preview=False, Color=p('02_hgcolor'), )
        n009 = mat.AddNode('ccl_SeparateRGB', name='Redfilter', loc=(1500,200), collapsed=False, preview=False )
        n010 = mat.AddNode('ccl_Bump', name='Bump2Normal', loc=(1500,400), collapsed=False, preview=False, Distance=BmpHgt,)
        n011 = mat.GetMap('color')
        n012 = mat.GetMap('bump')
        n019 = mat.GetMap('normal')
        n017 = mat.AddNode('image_map', name='SuntanMask', loc=(1500,600), collapsed=True, preview=True, Image_Source=p('02_suntanmask'), Filtering=4,)
        mat.GC(n017,True)
        n018 = mat.AddNode('ccl_Mix', name='ColorMix', loc=(1200,600), collapsed=False, preview=False, inputs=mydictionary)


        if n011 is None:
            n011=mat.AddNode('simple_color', name='Diffuse Color', loc=(700,20), collapsed=False, preview=False, )
        if n012 is None and n019 is None or p('02_procbump') is True:
            n013=mat.AddNode('ccl_VoronoiTexture', name='Pores', loc=(1800,20), collapsed=False, preview=False, Scale=1050, Coloring=2,)
            n014=mat.AddNode('ccl_Invert', name='Inverted Redfilter', loc=(1800,200), collapsed=False, preview=False,)
            n015=mat.AddNode('ccl_Bump', name='Make pores', loc=(1800,400), collapsed=False, preview=False, Distance=PrcBmpHgt, )
            n016=mat.AddNode('ccl_Math', name='Multiply', loc =(1800, 600), collapsed=False, preview=False, Type=20, )
        

        # Connect nodes
        mat.ConnectNodes(n001, n000, 'Surface')
        mat.ConnectNodes(n002, n001, 'Closure1')
        mat.ConnectNodes(n003, n001, 'Fac')
        mat.ConnectNodes(n004, n002, 'Closure1')
        mat.ConnectNodes(n005, n002, 'Closure2')
        mat.ConnectNodes(n006, n001, 'Closure2')
        mat.ConnectNodes(n018, n004, 'Color')
        mat.ConnectNodes(n007, n005, 'Color')
        mat.ConnectNodes(n008, n006, 'Color')
        mat.ConnectNodes(n009, n007, 'Value')
        mat.ConnectNodes(n009, n008, 'Value')
        mat.ConnectNodes(n009, n004, 'Texture Blur')
        mat.ConnectNodes(n011, n018, 'Color1')
        mat.ConnectNodes(n017, n018, 'Fac')        
        mat.ConnectNodes(n018, n009, 'Image')
        mat.ConnectNodes(n010, n003, 'Normal')
        mat.ConnectNodes(n010, n004, 'Normal')
        mat.ConnectNodes(n010, n005, 'Normal')
        mat.ConnectNodes(n010, n006, 'Normal')
        if n012 is None and n019 is None or p('02_procbump') is True:
                mat.ConnectNodes(n009, n014, 'Color')
                mat.ConnectNodes(n014, n010, 'Height')
                mat.ConnectNodes(n009, n016, 'Value1')
                mat.ConnectNodes(n013, n016, 'Value2')
                mat.ConnectNodes(n016, n015, 'Height')
                mat.ConnectNodes(n015, n010, 'Normal')
        else:
                if n012 :
                        mat.ConnectNodes(n012, n010, 'Height')
                elif n019:
                        if p('02_gradbump') is True :
                                mat.ConnectNodes(n019, n010, 'Height')
                        else: 
                                mat.ConnectNodes(n019, n010, 'Normal')
                                        
        # Wetaddon
        nw01 = None
        
        if p('02_wetskin') is True:
            nw01=self.WetSkin(mat, n000, n001)


        # Foamaddon
        

        if p('02_foam') is True:
            self.FoamySkin(mat, p, n000, n001, nw01)



        #Greasy

        if p('02_effects') == 1:
                self.GreasySkin(mat, p, n000, n001)

        # Yellow Custard Cover

        if p('02_effects') == 2:
                self.YellowCustard(mat, p, n000, n001, False)

        # Muddy

        if p('02_effects') == 3:
                self.Mud(mat, p, n000, n001, False)




    # Our Twelfth shader building method
    def CycleFace(self, mat, p):

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        #Convert Input parameters
        BmpHgt=p('02_bumpheight')/1000.0
        SSSDpt=p('02_sssdepth')/1000.0
        PrcBmpHgt=p('02_procbumpheight')/1000.0

        # Prepare Node
        n001 = mat.AddNode('ccl_MixClosure', name='MixClosure1', loc=(300,20), collapsed=False, preview=False, )
        n002 = mat.AddNode('ccl_MixClosure', name='MixClosure2', loc=(600,200), collapsed=False, preview=False, Fac=0.273)
        n003 = mat.AddNode('ccl_Fresnel', name='Fresnel1', loc=(600,20), collapsed=False, preview=False, IOR=1.45)
        n004 = mat.AddNode('ccl_SubsurfaceScattering', name='SSS', loc=(900,20), collapsed=False, preview=False, Scale=SSSDpt, Radius=(3.67,1.76,0.6), Falloff=1,)
        n005 = mat.AddNode('ccl_GlossyBsdf', name='Lowgloss', loc=(900,200), collapsed=False, preview=False, Roughness=p('02_lowgloss'), )
        n007 = mat.AddNode('ccl_HSV', name='LowglossColor', loc=(1200,200), collapsed=False, preview=False, Color=p('02_lgcolor'), )
        n006 = mat.AddNode('ccl_GlossyBsdf', name='Highgloss', loc=(900,400), collapsed=False, preview=False, Roughness=p('02_highgloss'), )
        n008 = mat.AddNode('ccl_HSV', name='HighglossColor', loc=(1200,400), collapsed=False, preview=False, Color=p('02_hgcolor'), )
        n009 = mat.AddNode('ccl_SeparateRGB', name='Redfilter', loc=(1500,200), collapsed=False, preview=False )
        n010 = mat.AddNode('ccl_Bump', name='Bump2Normal', loc=(1500,400), collapsed=False, preview=False, Distance=BmpHgt,)
        n011 = mat.GetMap('color')
        n012 = mat.GetMap('bump')
        n017 = mat.GetMap('normal')


        if n011 is None:
            n011=mat.AddNode('simple_color', name='Diffuse Color', loc=(700,20), collapsed=False, preview=False, )
        if n012 is None and n017 is None or p('02_procbump') is True:
            n013=mat.AddNode('ccl_VoronoiTexture', name='Pores', loc=(1800,20), collapsed=False, preview=False, Scale=1050, Coloring=2,)
            n014=mat.AddNode('ccl_Invert', name='Inverted Redfilter', loc=(1800,200), collapsed=False, preview=False,)
            n015=mat.AddNode('ccl_Bump', name='Make pores', loc=(1800,400), collapsed=False, preview=False, Distance=PrcBmpHgt, )
            n016=mat.AddNode('ccl_Math', name='Multiply', loc =(1800, 600), collapsed=False, preview=False, Type=20, )
        

        # Connect nodes
        mat.ConnectNodes(n001, n000, 'Surface')
        mat.ConnectNodes(n002, n001, 'Closure1')
        mat.ConnectNodes(n003, n001, 'Fac')
        mat.ConnectNodes(n004, n002, 'Closure1')
        mat.ConnectNodes(n005, n002, 'Closure2')
        mat.ConnectNodes(n006, n001, 'Closure2')
        mat.ConnectNodes(n011, n004, 'Color')
        mat.ConnectNodes(n007, n005, 'Color')
        mat.ConnectNodes(n008, n006, 'Color')
        mat.ConnectNodes(n009, n007, 'Value')
        mat.ConnectNodes(n009, n008, 'Value')
        mat.ConnectNodes(n009, n004, 'Texture Blur')        
        mat.ConnectNodes(n011, n009, 'Image')
        mat.ConnectNodes(n010, n003, 'Normal')
        mat.ConnectNodes(n010, n004, 'Normal')
        mat.ConnectNodes(n010, n005, 'Normal')
        mat.ConnectNodes(n010, n006, 'Normal')
        if n012 is None and n017 is None  or p('02_procbump') is True:
                mat.ConnectNodes(n009, n014, 'Color')
                mat.ConnectNodes(n014, n010, 'Height')
                mat.ConnectNodes(n009, n016, 'Value1')
                mat.ConnectNodes(n013, n016, 'Value2')
                mat.ConnectNodes(n016, n015, 'Height')
                mat.ConnectNodes(n015, n010, 'Normal')
        else:
                if n012 :
                        mat.ConnectNodes(n012, n010, 'Height')                  
                elif n017 :
                        if p('02_gradbump') is True :
                                mat.ConnectNodes(n017, n010, 'Height')
                        else: 
                                mat.ConnectNodes(n017, n010, 'Normal')
                                
        # Wetaddon
        nw01 = None
        
        if p('02_wetskin') is True and p('02_nowetface') is False:
            nw01=self.WetSkin(mat, n000, n001)


        # Foamaddon
        

        if p('02_foam') is True and p('02_nofoamyface') is False:
            self.FoamySkin(mat, p, n000, n001, nw01)

                        
        #Greasy

        if p('02_effects') == 1:
                self.GreasySkin(mat, p, n000, n001)

        # Yellow Custard Cover

        if p('02_effects') == 2:
                self.YellowCustard(mat, p, n000, n001, False)

        # Muddy

        if p('02_effects') == 3:
                self.Mud(mat, p, n000, n001, False)


    # Our Thirteenth shader building method
    def CycleHair(self, mat, p):
                

        # Prepare the root node
        n000 = self._PrepareRoot(mat, p)

        #Prepare Nodes
        n001=mat.AddNode('ccl_MixClosure', name = 'HairClosure1', loc=(200,100), collapsed=False, preview=False,)
        n002=mat.AddNode('ccl_MixClosure', name = 'HairClosure2', loc=(450,300), collapsed=False, preview=False, )
        n002a=mat.AddNode('ccl_Fresnel', name = 'HairFresnel', loc=(450,0),collapsed=False, preview=False, IOR = 1.5,)
        n003=mat.AddNode('ccl_TransparentBsdf', name = 'Transparant', loc=(450,100), collapsed=False, preview=False, Color =(1.0,1.0,1.0),)
        n004=mat.AddNode('ccl_DiffuseBsdf', name = 'Haircolor', loc=(700,100), collapsed=False, preview=False, Roughness = 0.3,)
        n005=mat.AddNode('ccl_AnisotropicBsdf', name = 'Hairshine', loc=(700,300), collapsed=False, preview=False, Color = p('02_hairshinecolor'), Roughness = p('02_hairroughness'), Anisotropy=0.5,)
        n006=mat.AddNode('ccl_Bump', name = 'Hairbump', loc=(950,200), collapsed=False, preview=False, Distance=0.002,)
        n007=mat.GetMap('color')
        n008=mat.GetMap('transparency')
        n009=mat.GetMap('bump')
        mat.GC(n009,True)

        if n007 is None :
            n007=mat.AddNode('simple_color', name='Diffuse Color', loc=(700,20), collapsed=False, preview=False, Color = (0.0,0.0,0.0) )

        if n009 is None :
            n010=mat.AddNode('ccl_HSV', name = 'Color2Grey', loc=(950,450), collapsed=False, preview=False, Saturation = 0.0)

        #ConnectNodes

        mat.ConnectNodes(n001, n000, 'Surface')
        mat.ConnectNodes(n008, n001, 'Fac')
        mat.ConnectNodes(n003, n001, 'Closure1')
        mat.ConnectNodes(n002, n001, 'Closure2')
        mat.ConnectNodes(n002a, n002, 'Fac')
        mat.ConnectNodes(n004, n002, 'Closure1')
        mat.ConnectNodes(n005, n002, 'Closure2')
        mat.ConnectNodes(n006, n004, 'Normal')
        mat.ConnectNodes(n006, n005, 'Normal')
        mat.ConnectNodes(n006, n002a, 'Normal')
        mat.ConnectNodes(n007, n004, 'Color')

        if n009 is None :
                mat.ConnectNodes(n007, n010, 'Color')
                mat.ConnectNodes(n010, n006, 'Height')
        else :
                mat.ConnectNodes(n009, n006, 'Height')
        
        # Yellow Custard Cover

        if p('02_effects') == 2:
                self.YellowCustard(mat, p, n001, n002, True)

        # Muddy

        if p('02_effects') == 3:
                self.Mud(mat, p, n001, n002, True)                                 


    def Renderer(self):
        # Return the render engine this plugin is designed to work with
        # Must be 'Firefly' or 'Superfly'
        return 'Superfly'



